
# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Defines the Cases entity.
"""
from __future__ import unicode_literals, print_function
import os
import sys
import importlib

import numpy as np

from HERON.src.base import Base

from HERON.src.dispatch.Factory import known as known_dispatchers
from HERON.src.dispatch.Factory import get_class as get_dispatcher

from HERON.src.ValuedParams import factory as vp_factory
from HERON.src.ValuedParamHandler import ValuedParamHandler

from HERON.src.validators.Factory import known as known_validators
from HERON.src.validators.Factory import get_class as get_validator

import HERON.src._utils as hutils
try:
  import ravenframework
except ModuleNotFoundError:
  framework_path = hutils.get_raven_loc()
  sys.path.append(framework_path)
from ravenframework.utils import InputData, InputTypes

class Case(Base):
  """
    Produces something, often as the cost of something else
    TODO this case is for "sweep-opt", need to make a superclass for generic
  """

  # metrics that can be used for objective in optimization or returned with results
  # each metric contains a dictionary with the following keys:
  # 'prefix' - printed result name
  # 'optimization_default' - 'min' or 'max' for optimization
  # 'percent' (only for percentile) - list of percentiles to return
  # 'threshold' (only for sortinoRatio, gainLossRatio, expectedShortfall, valueAtRisk) - threshold value for calculation
  metrics_mapping = {'expectedValue': {'prefix': 'mean', 'optimization_default': 'max'},
                     'minimum': {'prefix': 'min', 'optimization_default': 'max'},
                     'maximum': {'prefix': 'max', 'optimization_default': 'max'},
                     'median': {'prefix': 'med', 'optimization_default': 'max'},
                     'variance': {'prefix': 'var', 'optimization_default': 'min'},
                     'sigma': {'prefix': 'std', 'optimization_default': 'min'},
                     'percentile': {'prefix': 'perc', 'optimization_default': 'max', 'percent': ['5', '95']},
                     'variationCoefficient': {'prefix': 'varCoeff', 'optimization_default': 'min'},
                     'skewness': {'prefix': 'skew', 'optimization_default': 'min'},
                     'kurtosis': {'prefix': 'kurt', 'optimization_default': 'min'},
                     'samples': {'prefix': 'samp'},
                     'sharpeRatio': {'prefix': 'sharpe', 'optimization_default': 'max'},
                     'sortinoRatio': {'prefix': 'sortino', 'optimization_default': 'max', 'threshold': 'median'},
                     'gainLossRatio': {'prefix': 'glr', 'optimization_default': 'max', 'threshold': 'median'},
                     'expectedShortfall': {'prefix': 'es', 'optimization_default': 'min', 'threshold': ['0.05']},
                     'valueAtRisk': {'prefix': 'VaR', 'optimization_default': 'min', 'threshold': ['0.05']}}

  #### INITIALIZATION ####
  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, input_specs, InputData, specs
    """
    input_specs = InputData.parameterInputFactory('Case', ordered=False, baseNode=None,
                                                  descr=r"""The \xmlNode{Case} node contains the general physics and
                                                  economics information required for a HERON workflow to be created
                                                  and solved.""")
    input_specs.addParam('name', param_type=InputTypes.StringType, required=True,
                         descr=r"""the name by which this analysis should be referred within HERON.""")

    # Optional Identifier Nodes
    label_specs = InputData.parameterInputFactory(name='label', ordered=False,
                                                  descr=r"""provides static label information to the model;
                                                  unused in computation. These data will be passed along through
                                                  the meta class and output in the simulation result files.
                                                  These data can also be accessed within user-defined transfer
                                                  functions by using \texttt{meta['HERON']['Case'].get_labels()}.""")
    label_specs.addParam(name='name',param_type=InputTypes.StringType,
                         descr=r"""the generalized name of the identifier.
                         Example: ``$<$label name="state"$>$Idaho$<$/label$>$''""")
    input_specs.addSub(label_specs)

    mode_options = InputTypes.makeEnumType('ModeOptions', 'ModeOptionsType', ['opt', 'sweep'])
    desc_mode_options = r"""determines the mode of operation for the outer/inner RAVEN.
                        If ``sweep'' then parametrically sweep over distributed values.
                        If ``opt'' then search distributed values for economic metric optima.
                        """
    input_specs.addSub(InputData.parameterInputFactory('mode', contentType=mode_options, strictMode=True,
                                                       descr=desc_mode_options))

    verbosity_options = InputTypes.makeEnumType('VerbosityOptions', 'VerbosityOptionsType',
                                                ['silent', 'quiet', 'all', 'debug'])
    desc_verbosity_options = r"""determines the level of verbosity for the outer/inner RAVEN runs. \default{all}.
                             If ``silent'' only errors are displayed.
                             If ``quiet'' errors and warnings are displayed.
                             If ``all'' errors, warnings, and messages are displayed.
                             If ``debug'' errors, warnings, messages, and debug messages are displayed."""
    input_specs.addSub(InputData.parameterInputFactory('verbosity', contentType=verbosity_options,
                                                       strictMode=True, descr=desc_verbosity_options))

    workflow_options = InputTypes.makeEnumType('WorkflowOptions', 'WorkflowOptionsType',
                                               ['standard', 'MOPED', 'combined', 'DISPATCHES'])

    desc_workflow_options = r"""determines the desired workflow(s) for the HERON analysis. \default{standard}.
                            If ``standard'' runs HERON as usual (writes outer/inner for RAVEN workflow).
                            If ``MOPED'' runs monolithic solver MOPED using the information in xml input.
                            If ``combined'' runs both workflows, setting up RAVEN workflow and solving with MOPED.
                            See Workflow Options section in user guide for more details"""
    input_specs.addSub(InputData.parameterInputFactory('workflow', contentType=workflow_options,
                                                       strictMode=True, descr=desc_workflow_options))

    # not yet implemented TODO
    #econ_metrics = InputTypes.makeEnumType('EconMetrics', 'EconMetricsTypes', ['NPV', 'lcoe'])
    #desc_econ_metrics = r"""indicates the economic metric that should be used for the HERON analysis. For most cases, this
    #                    should be NPV."""
    # input_specs.addSub(InputData.parameterInputFactory('metric', contentType=econ_metrics, descr=desc_econ_metrics))
    # input_specs.addSub(InputData.parameterInputFactory('differential', contentType=InputTypes.BoolType, strictMode=True,
    # descr=r"""(not implemented) allows differentiation between two HERON runs as a desired
    # economic metric."""

    # debug mode, for checking dispatch and etc
    debug = InputData.parameterInputFactory('debug', descr=r"""Including this node enables a reduced-size
        run with increased outputs for checking how the sampling, dispatching, and cashflow mechanics
        are working for a particular input. Various options for modifying how the debug mode operates
        are included for convenience; however, just including this node will result in a minimal run.""")
    debug.addSub(InputData.parameterInputFactory('inner_samples', contentType=InputTypes.IntegerType,
        descr=r"""sets the number of inner realizations of the stochastic synthetic histories and dispatch
              optimization to run per outer sample. Overrides the \xmlNode{num_arma_steps} option while
              \xmlNode{debug} mode is enabled. \default{1}"""))
    debug.addSub(InputData.parameterInputFactory('macro_steps', contentType=InputTypes.IntegerType,
        descr=r"""sets the number of macro steps (e.g. years) the stochastic synthetic histories and dispatch
              optimization should include. \default{1}"""))
    debug.addSub(InputData.parameterInputFactory('dispatch_plot', contentType=InputTypes.BoolType,
        descr=r"""provides a dispatch plot after running through \xmlNode{inner_samples} and
              \xmlNode{macro_steps} provided. To prevent plotting output during debug mode set to "False".
              \default{True}"""))
    input_specs.addSub(debug)

    parallel = InputData.parameterInputFactory('parallel', descr=r"""Describes how to parallelize this run. If not present defaults to no parallelization (1 outer, 1 inner)""")
    parallel.addSub(InputData.parameterInputFactory('outer', contentType=InputTypes.IntegerType,
        descr=r"""the number of parallel runs to use for the outer optimization run. The product of this
              number and \xmlNode{inner} should be at most the number of parallel process available on
              your computing device. This should also be at most the number of samples needed per outer iteration;
              for example, with 3 opt bound variables and using finite differencing, at most 4 parallel outer runs
              can be used. \default{number of variable sweeps + 1}"""))
    parallel.addSub(InputData.parameterInputFactory('inner', contentType=InputTypes.IntegerType,
        descr=r"""the number of parallel runs to use per inner sampling run. This should be at most the number
              of denoising samples, and at most the number of parallel processes available on your computing
              device. \default{number of denoising samples}"""))
    #XXX RAVEN should be providing this InputData
    runinfo = InputData.parameterInputFactory('runinfo',
                descr=r"""this is copied into the RAVEN runinfo block, and defaults are specified in RAVEN""")
    runinfo.addSub(InputData.parameterInputFactory('expectedTime', contentType=InputTypes.StringType,
                  descr=r"""the expected time for the run to take in hours, minutes, seconds (example 24:00:00 for 1 day) """))
    runinfo.addSub(InputData.parameterInputFactory('clusterParameters', contentType=InputTypes.StringType,
                  descr=r"""Extra parameters needed by the cluster qsub command"""))
    runinfo.addSub(InputData.parameterInputFactory('RemoteRunCommand', contentType=InputTypes.StringType,
                   descr=r"""The shell command used to run remote commands"""))
    runinfo.addSub(InputData.parameterInputFactory('memory', contentType=InputTypes.StringType,descr=r"""The amount of memory needed per core (example 4gb)"""))
    parallel.addSub(runinfo)
    # TODO HPC?
    input_specs.addSub(parallel)

    data_handling = InputData.parameterInputFactory('data_handling', descr=r"""Provides options for data handling within HERON operations.""")
    inner_outer_data = InputTypes.makeEnumType('InnerOuterData', 'InnerOuterDataType', ['csv', 'netcdf'])
    data_handling.addSub(InputData.parameterInputFactory('inner_to_outer', contentType=inner_outer_data,
        descr=r"""which type of data format to transfer results from inner (stochastic dispatch optimization) runs to
                  the outer (capacity and meta-variable optimization) run. CSV is generally slower and not recommended,
                  but may be useful for debugging. NetCDF is more generally more efficient. \default{netcdf}"""))
    input_specs.addSub(data_handling)

    input_specs.addSub(InputData.parameterInputFactory('num_arma_samples', contentType=InputTypes.IntegerType,
                                                       descr=r"""provides the number of synthetic histories that should
                                                       be considered per system configuration in order to obtain a
                                                       reasonable representation of the economic metric. Sometimes
                                                       referred to as ``inner samples'' or ``denoisings''."""))

    # time discretization
    time_discr = InputData.parameterInputFactory('time_discretization',
                                                 descr=r"""node that defines how within-cycle time discretization should
                                                 be handled for solving the dispatch.""")
    time_discr.addSub(InputData.parameterInputFactory('time_variable', contentType=InputTypes.StringType,
                                                      descr=r"""name for the \texttt{time} variable used in this
                                                      simulation. \default{time}"""))
    time_discr.addSub(InputData.parameterInputFactory('year_variable', contentType=InputTypes.StringType,
                                                      descr=r"""name for the \texttt{year} or \texttt{macro} variable
                                                      used in this simulation. \default{Year}"""))
    time_discr.addSub(InputData.parameterInputFactory('start_time', contentType=InputTypes.FloatType,
                                                      descr=r"""value for \texttt{time} variable at which the inner
                                                      dispatch should begin. \default{0}"""))
    time_discr.addSub(InputData.parameterInputFactory('end_time', contentType=InputTypes.FloatType,
                                                      descr=r"""value for \texttt{time} variable at which the inner
                                                      dispatch should end. If not specified, both \xmlNode{time_interval}
                                                      and \xmlNode{num_timesteps} must be defined."""))
    time_discr.addSub(InputData.parameterInputFactory('num_steps', contentType=InputTypes.IntegerType,
                                                      descr=r"""number of discrete time steps for the inner dispatch.
                                                      Either this node or \xmlNode{time_interval} must be defined."""))
    time_discr.addSub(InputData.parameterInputFactory('time_interval',contentType=InputTypes.FloatType,
                                                      descr=r"""length of a time step for the inner dispatch, in units of
                                                      the time variable (not indices). Either this node or
                                                      \xmlNode{num_timesteps} must be defined. Note that if an integer
                                                      number of intervals do not fit between \xmlNode{start_time} and
                                                      \xmlNode{end_time}, an error will be raised."""))
    input_specs.addSub(time_discr)

    # economics global settings
    econ = InputData.parameterInputFactory('economics', ordered=False,
                                           descr= r"""node containing general economic setting in which to perform
                                           HERON analysis.""")
    econ.addSub(InputData.parameterInputFactory('ProjectTime', contentType=InputTypes.FloatType,
                                                descr=r"""the number of cycles (usually years) for the HERON analysis
                                                to cover."""))
    econ.addSub(InputData.parameterInputFactory('DiscountRate', contentType=InputTypes.FloatType,
                                                descr=r"""rate representing the time value of money to the firm used
                                                to discount cash flows in the multicycle economic analysis. Passed to
                                                the CashFlow module."""))
    econ.addSub(InputData.parameterInputFactory('tax', contentType=InputTypes.FloatType,
                                                descr=r"""the taxation rate, a metric which represents the
                                                rate at which the firm is taxed. Passed to the CashFlow module."""))
    econ.addSub(InputData.parameterInputFactory('inflation', contentType=InputTypes.FloatType,
                                                descr=r"""a metric which represents the rate at which the average
                                                price of goods and services in an economy increases over a cycle,
                                                usually a year. Passed to the CashFlow module."""))
    econ.addSub(InputData.parameterInputFactory('verbosity', contentType=InputTypes.IntegerType,
                                                descr=r"""the level of output to print from the CashFlow calculations.
                                                Passed to the CashFlow module."""))
    # is this actually CashFlow verbosity or is it really HERON verbosity?
    input_specs.addSub(econ)

    # dispatcher
    dispatch = InputData.parameterInputFactory('dispatcher', ordered=False,
                                               descr=r"""This node defines the dispatch strategy and options to use in
                                               the ``inner'' run.""")
    for d in known_dispatchers:
      vld_spec = get_dispatcher(d).get_input_specs()
      dispatch.addSub(vld_spec)
    input_specs.addSub(dispatch)

    # validator
    validator = InputData.parameterInputFactory('validator', ordered=False,
                                                descr=r"""This node defines the dispatch validation strategy and options
                                                to use in the ``inner'' run.""")
    for d in known_validators:
      vld_spec = get_validator(d).get_input_specs()
      validator.addSub(vld_spec)
    input_specs.addSub(validator)

    # optimization settings
    optimizer = InputData.parameterInputFactory('optimization_settings',
                                                descr=r"""This node defines the settings to be used for the optimizer in
                                                the ``outer'' run.""")
    metric_options = InputTypes.makeEnumType('MetricOptions', 'MetricOptionsType', list(cls.metrics_mapping.keys()))
    desc_metric_options = r"""determines the statistical metric (calculated by RAVEN BasicStatistics
                          or EconomicRatio PostProcessors) from the ``inner'' run to be used as the
                          objective in the ``outer'' optimization.
                          \begin{itemize}
                            \item For ``percentile'' the additional parameter \textit{percent=`X'}
                            is required where \textit{X} is the requested percentile (a floating
                            point value between 0.0 and 100.0).
                            \item For ``sortinoRatio'' and ``gainLossRatio'' the additional
                            parameter \textit{threshold=`X'} is required where \textit{X} is the
                            requested threshold (`median' or `zero').
                            \item For ``expectedShortfall'' and ``valueAtRisk'' the additional
                            parameter \textit{threshold=`X'} is required where \textit{X} is the
                            requested $\alpha$ value (a floating point value between 0.0 and 1.0).
                          \end{itemize}
                          """
    metric = InputData.parameterInputFactory('metric', contentType=metric_options, strictMode=True,
                                             descr=desc_metric_options)
    metric.addParam(name='percent',
                    param_type=InputTypes.FloatType,
                    descr=r"""requested percentile (a floating point value between 0.0 and 100.0).
                              Required when \xmlNode{metric} is ``percentile.''
                              \default{5}""")
    metric.addParam(name='threshold',
                    param_type=InputTypes.StringType,
                    descr=r"""\begin{itemize}
                                \item requested threshold (`median' or `zero'). Required when
                                \xmlNode{metric} is ``sortinoRatio'' or ``gainLossRatio.''
                                \default{`zero'}
                                \item requested $ \alpha $ value (a floating point value between 0.0
                                and 1.0). Required when \xmlNode{metric} is ``expectedShortfall'' or
                                ``valueAtRisk.'' \default{0.05}
                              \end{itemize}""")
    optimizer.addSub(metric)
    type_options = InputTypes.makeEnumType('TypeOptions', 'TypeOptionsType',
                                           ['min', 'max'])
    desc_type_options = r"""determines whether the objective should be minimized or maximized.
                            \begin{itemize}
                              \item when metric is ``expectedValue,'' ``minimum,'' ``maximum,''
                              ``median,'' ``percentile,'' ``sharpeRatio,'' ``sortinoRatio,''
                              ``gainLossRatio'' \default{max}
                              \item when metric is ``variance,'' ``sigma,'' ``variationCoefficient,''
                              ``skewness,'' ``kurtosis,'' ``expectedShortfall,'' ``valueAtRisk''
                              \default{min}
                            \end{itemize}"""
    type_sub = InputData.parameterInputFactory('type', contentType=type_options, strictMode=True,
                                               descr=desc_type_options)
    optimizer.addSub(type_sub)
    persistenceSub = InputData.parameterInputFactory('persistence',contentType=InputTypes.IntegerType,
                                                      descr=r"""provides the number of consecutive times convergence should be reached before a trajectory
                                                      is considered fully converged. This helps in preventing early false convergence.""" )
    optimizer.addSub(persistenceSub)
    input_specs.addSub(optimizer)

    convergence = InputData.parameterInputFactory('convergence',
                                                  descr=r"""defines the optimization convergence criteria.""")
    gradient_sub = InputData.parameterInputFactory('gradient',
                                                    descr=r"""termination criterion for the gradient
                                                              \default{1e-4}""")
    convergence.addSub(gradient_sub)
    objective_sub = InputData.parameterInputFactory('objective',
                                                    descr=r"""termination criterion for the objective function
                                                              \default{1e-8}""")
    convergence.addSub(objective_sub)
    stepsize_sub = InputData.parameterInputFactory('stepSize',
                                                    descr=r"""termination criterion for the design space step size""")
    convergence.addSub(stepsize_sub)

    optimizer.addSub(convergence)

    # Add magic variables that will be passed to the outer and inner.
    dispatch_vars = InputData.parameterInputFactory(
        'dispatch_vars',
        descr=r"This node defines a set containing additional variables"
        "to sample that are not associated with a specific component."
    )
    value_param = vp_factory.make_input_specs(
        'variable',
        descr=r"This node defines the single additional dispatch variable used in the case."
    )
    value_param.addParam(
        'name',
        param_type=InputTypes.StringType,
        descr=r"The unique name of the dispatch variable."
    )
    dispatch_vars.addSub(value_param)
    input_specs.addSub(dispatch_vars)

    # result statistics
    result_stats = InputData.parameterInputFactory('result_statistics',
                                                   descr=r"""This node defines the additional statistics
                                                   to be returned with the results. The statistics
                                                   \texttt{expectedValue} (prefix ``mean''),
                                                   \texttt{sigma} (prefix ``std''), and \texttt{median}
                                                   (prefix ``med'') are always returned with the results.
                                                   Each subnode is the RAVEN-style name of the desired
                                                   return statistic.""")
    for stat in cls.metrics_mapping:
      if stat not in ['expectedValue', 'sigma', 'median']:
        statistic = InputData.parameterInputFactory(stat, strictMode=True,
                                                    descr=r"""{} uses the prefix ``{}'' in the result output.""".format(stat, cls.metrics_mapping[stat]['prefix']))
        if stat == 'percentile':
          statistic.addParam('percent', param_type=InputTypes.StringType,
                            descr=r"""requested percentile (a floating point value between 0.0 and 100.0).
                            When no percent is given, returns both 5th and 95th percentiles.""")
        elif stat in ['sortinoRatio', 'gainLossRatio']:
          statistic.addParam('threshold', param_type=InputTypes.StringType,
                            descr=r"""requested threshold (``median" or ``zero").""", default='``median"')
        elif stat in ['expectedShortfall', 'valueAtRisk']:
          statistic.addParam('threshold', param_type=InputTypes.StringType,
                            descr=r"""requested threshold (a floating point value between 0.0 and 1.0).""",
                            default='``0.05"')
        result_stats.addSub(statistic)
    input_specs.addSub(result_stats)

    return input_specs

  def __init__(self, run_dir, **kwargs):
    """
      Constructor
      @ In, None
      @ Out, None
    """
    Base.__init__(self, **kwargs)
    self.name = None                   # case name
    self._mode = None                  # extrema to find: opt, sweep
    self._metric = 'NPV'               # TODO: future work - economic metric to focus on: lcoe, profit, cost
    self.run_dir = run_dir             # location of HERON input file
    self._verbosity = 'all'            # default verbosity for RAVEN inner/outer

    self.dispatch_name = None          # name of dispatcher to use
    self.dispatcher = None             # type of dispatcher to use
    self.validator_name = None         # name of dispatch validation to use
    self.validator = None              # type of dispatch validation to use
    self.dispatch_vars = {}            # non-component optimization ValuedParams

    self.useParallel = False           # parallel tag specified?
    self.outerParallel = 0             # number of outer parallel runs to use
    self.innerParallel = 0             # number of inner parallel runs to use

    self._diff_study = None            # is this only a differential study?
    self._num_samples = 1              # number of ARMA stochastic samples to use ("denoises")
    self._hist_interval = None         # time step interval, time between production points
    self._hist_len = None              # total history length, in same units as _hist_interval
    self._num_hist = None              # number of history steps, hist_len / hist_interval
    self._global_econ = {}             # global economics settings, as a pass-through
    self._increments = {}              # stepwise increments for resource balancing
    self._time_varname = 'time'        # name of the time-variable throughout simulation
    self._year_varname = 'Year'        # name of the year-variable throughout simulation
    self._labels = {}                  # extra information pertaining to current case
    self.debug = {                     # debug options, as enabled by the user (defaults included)
        'enabled': False,              # whether to enable debug mode
        'inner_samples': 1,            # how many inner realizations to sample
        'macro_steps': 1,              # how many "years" for inner realizations
        'dispatch_plot': True          # whether to output a plot in debug mode
    }

    self.data_handling = {             # data handling options
      'inner_to_outer': 'netcdf',      # how to pass inner data to outer (csv, netcdf)
    }

    self._time_discretization = None   # (start, end, number) for constructing time discretization, same as argument to np.linspace
    self._Resample_T = None            # user-set increments for resources
    self._optimization_settings = None # optimization settings dictionary for outer optimization loop
    self._workflow = 'standard' # setting for how to run HERON, default is through raven workflow
    self._result_statistics = {        # desired result statistics (keys) dictionary with attributes (values)
        'sigma': None,                 # user can specify additional result statistics
        'expectedValue': None,
        'median': None}

    # clean up location
    self.run_dir = os.path.abspath(os.path.expanduser(self.run_dir))

  def read_input(self, xml):
    """
      Sets settings from input file
      @ In, xml, xml.etree.ElementTree.Element, input from user
      @ Out, None
    """
    # get specs for allowable inputs
    specs = self.get_input_specs()()
    specs.parseNode(xml)
    self.name = specs.parameterValues['name']
    for item in specs.subparts:
      # TODO move from iterative list to seeking list, at least for required nodes?
      if item.getName() == 'debug':
        self.debug['enabled'] = True
        for node in item.subparts:
          self.debug[node.getName()] = node.value
      elif item.getName() == 'label':
        self._labels[item.parameterValues['name']] = item.value
      if item.getName() == 'verbosity':
        self._verbosity = item.value
      if item.getName() == 'mode':
        self._mode = item.value
      elif item.getName() == 'parallel':
        self.useParallel = True
        self.parallelRunInfo = {}
        for sub in item.subparts:
          if sub.getName() == 'outer':
            self.outerParallel = sub.value
          elif sub.getName() == 'inner':
            self.innerParallel = sub.value
          elif sub.getName() == 'runinfo':
            for subsub in sub.subparts:
              self.parallelRunInfo[subsub.getName()] = str(subsub.value)
      elif item.getName() == 'metric':
        self._metric = item.value
      elif item.getName() == 'differential':
        self._diff_study = item.value
      elif item.getName() == 'num_arma_samples':
        self._num_samples = item.value
      elif item.getName() == 'time_discretization':
        self._time_discretization = self._read_time_discr(item)
      elif item.getName() == 'economics':
        for sub in item.subparts:
          self._global_econ[sub.getName()] = sub.value
          if self.debug['enabled'] and sub.getName() == "ProjectTime":
            self._global_econ[sub.getName()] = self.debug['macro_steps']
      elif item.getName() == 'dispatcher':
        # instantiate a dispatcher object.
        inp = item.subparts[0]
        name = inp.getName()
        typ = get_dispatcher(name)
        self.dispatcher = typ()
        self.dispatcher.read_input(inp)
      elif item.getName() == 'validator':
        vld = item.subparts[0]
        name = vld.getName()
        typ = get_validator(name)
        self.validator = typ()
        self.validator.read_input(vld)
      elif item.getName() == 'optimization_settings':
        self._optimization_settings = self._read_optimization_settings(item)
      elif item.getName() == 'dispatch_vars':
        for node in item.subparts:
          var_name = node.parameterValues['name']
          vp = ValuedParamHandler(var_name)
          vp.read(var_name, node, self.get_mode())
          self.dispatch_vars[var_name] = vp
      elif item.getName() == 'data_handling':
        self.data_handling = self._read_data_handling(item)
      elif item.getName() == 'workflow':
        self._workflow = item.value
      elif item.getName() == 'result_statistics':
        new_result_statistics = self._read_result_statistics(item)
        self._result_statistics.update(new_result_statistics)

    # checks
    if self._mode is None:
      self.raiseAnError('No <mode> node was provided in the <Case> node!')
    if self.dispatcher is None:
      self.raiseAnError('No <dispatch> node was provided in the <Case> node!')
    if self._time_discretization is None:
      self.raiseAnError('<time_discretization> node was not provided in the <Case> node!')
    if self.innerParallel == 0 and self.useParallel:
      #set default inner parallel to number of samples (denoises)
      self.innerParallel = self._num_samples
    #Note that if self.outerParallel == 0 and self.useParallel
    # then outerParallel will be set in template_driver _modify_outer_samplers
    cores_requested = self.innerParallel * self.outerParallel
    if cores_requested > 1:
      # check to see if the number of processes available can meet the request
      detected = os.cpu_count() - 1 # -1 to prevent machine OS locking up
      if detected < cores_requested:
        self.raiseAWarning('System may be overloaded and greatly increase run time! ' +
                           f'Number of available cores detected: {detected}; ' +
                           f'Number requested: {cores_requested} (inner: {self.innerParallel} * outer: {self.outerParallel}) ')

    # TODO what if time discretization not provided yet?
    self.dispatcher.set_time_discr(self._time_discretization)
    self.dispatcher.set_validator(self.validator)

    self.raiseADebug(f'Successfully initialized Case {self.name}.')

  def _read_data_handling(self, node):
    """
      Reads the data handling node.
      @ In, node, InputParams.ParameterInput, data handling head node
      @ Out, settings, dict, options for data handling
    """
    settings = {}
    # read settings
    for sub in node.subparts:
      name = sub.getName()
      if name == 'inner_to_outer':
        settings['inner_to_outer'] = sub.value
    # set defaults
    if 'inner_to_outer' not in settings:
      settings['inner_to_outer'] = 'netcdf'
    return settings

  def _read_time_discr(self, node):
    """
      Reads the time discretization node.
      @ In, node, InputParams.ParameterInput, time discretization head node
      @ Out, discr, tuple, (start, end, num_steps) for creating numpy linspace
    """
    # name of time variable
    var_name = node.findFirst('time_variable')
    if var_name is not None:
      self._time_varname = var_name.value
    # name of year variable
    year_name = node.findFirst('year_variable')
    if year_name is not None:
      self._year_varname = year_name.value
    # start
    start_node = node.findFirst('start_time')
    if start_node is None:
      start = 0.0
    else:
      start = start_node.value
    # options:
    end_node = node.findFirst('end_time')
    dt_node = node.findFirst('time_interval')
    num_node = node.findFirst('num_steps')
    # - specify end and num steps
    if (end_node and num_node):
      end = end_node.value
      num = num_node.value
    # - specify end and dt
    elif (end_node and dt_node):
      end = end_node.value
      dt = dt_node.value
      num = int(np.floor((end - start) / dt))
    # - specify dt and num steps
    elif (dt_node and num_node):
      dt = dt_node.value
      num = num_node.value
      end = dt * num + start
    else:
      self.raiseAnError(IOError, 'Invalid time discretization choices! Must specify any of the following pairs: ' +
                                 '(<end_time> and <num_steps>) or ' +
                                 '(<end_time> and <time_interval>) or ' +
                                 '(<num_steps> and <time_interval>.)')
    # TODO can we take it automatically from an ARMA later, either by default or if told to?
    return (start, end, num)

  def _read_optimization_settings(self, node):
    """
      Reads optimization settings node
      @ In, node, InputParams.ParameterInput, optimization settings head node
      @ Out, opt_settings, dict, optimization settings as dictionary
    """
    opt_settings = {}
    for sub in node.subparts:
      sub_name = sub.getName()
      # add metric information to opt_settings dictionary
      if sub_name == 'metric':
        opt_settings[sub_name] = {}
        metric_name = sub.value
        opt_settings[sub_name]['name'] = metric_name
        # some metrics have an associated parameter
        if metric_name == 'percentile':
          try:
            opt_settings[sub_name]['percent'] = sub.parameterValues['percent']
          except KeyError:
            opt_settings[sub_name]['percent'] = 5
        elif metric_name in ['sortinoRatio', 'gainLossRatio']:
          try:
            opt_settings[sub_name]['threshold'] = sub.parameterValues['threshold']
          except KeyError:
            opt_settings[sub_name]['threshold'] = 'zero'
        elif metric_name in ['expectedShortfall', 'valueAtRisk']:
          try:
            opt_settings[sub_name]['threshold'] = sub.parameterValues['threshold']
          except KeyError:
            opt_settings[sub_name]['threshold'] = 0.05
      elif sub_name == 'convergence':
        opt_settings[sub_name] = {}
        for ssub in sub.subparts:
          opt_settings[sub_name][ssub.getName()] = ssub.value
      else:
        # add other information to opt_settings dictionary (type is only information implemented)
        opt_settings[sub_name] = sub.value

    return opt_settings

  def _read_result_statistics(self, node):
    """
      Reads result statistics node
      @ In, node, InputParams.ParameterInput, result statistics head node
      @ Out, result_statistics, dict, result statistics settings as dictionary
    """
    # result_statistics keys are statistic name value is percent, threshold value, or None
    result_statistics = {}
    for sub in node.subparts:
      sub_name = sub.getName()
      if sub_name == 'percentile':
        try:
          percent = sub.parameterValues['percent']
          # if multiple percents are given, set as a list
          if sub_name in result_statistics:
            if isinstance(result_statistics[sub_name], list):
              if percent not in result_statistics[sub_name]:
                result_statistics[sub_name].append(percent)
            else:
              result_statistics[sub_name] = [result_statistics[sub_name], percent]
          else:
            result_statistics[sub_name] = percent
        except KeyError:
          result_statistics[sub_name] = self.metrics_mapping[sub_name]['percent']
      elif sub_name in ['sortinoRatio', 'gainLossRatio']:
        try:
          result_statistics[sub_name] = sub.parameterValues['threshold']
        except KeyError:
          result_statistics[sub_name] = self.metrics_mapping[sub_name]['threshold']
      elif sub_name in ['expectedShortfall', 'valueAtRisk']:
        try:
          threshold = sub.parameterValues['threshold']
          # if multiple thresholds are given, set as a list
          if sub_name in result_statistics:
            if isinstance(result_statistics[sub_name], list):
              if threshold not in result_statistics[sub_name]:
                result_statistics[sub_name].append(threshold)
            else:
              result_statistics[sub_name] = [result_statistics[sub_name], threshold]
          else:
            result_statistics[sub_name] = sub.parameterValues['threshold']
        except KeyError:
          result_statistics[sub_name] = self.metrics_mapping[sub_name]['threshold']
      else:
        result_statistics[sub_name] = None

    return result_statistics

  def initialize(self, components, sources):
    """
      Called after all objects are created, allows post-input initialization
      @ In, components, list, HERON components
      @ In, sources, list, HERON sources (placeholders)
      @ Out, None
    """
    # check sources
    for src in sources:
      src.checkValid(self, components, sources)
    # dispatcher
    self.dispatcher.initialize(self, components, sources)

  def __repr__(self):
    """
      Determines how this class appears when printed.
      @ In, None
      @ Out, repr, str, string representation
    """
    return '<HERON Case>'

  def print_me(self, tabs=0, tab='  ', **kwargs):
    """
      Prints info about self
      @ In, tabs, int, number of tabs to insert before print
      @ In, tab, str, tab prefix
      @ Out, None
    """
    pre = tab*tabs
    self.raiseADebug(pre+'Case:')
    self.raiseADebug(pre+'  name:', self.name)
    self.raiseADebug(pre+'  mode:', self._mode)
    self.raiseADebug(pre+'  meric:', self._metric)
    self.raiseADebug(pre+'  diff_study:', self._diff_study)

  #### ACCESSORS ####
  def get_increments(self):
    """
      Accessor.
      @ In, None
      @ Out, self._increments, dict, increments for resource evaluation
    """
    return self._increments

  def get_working_dir(self, which):
    """
      Accessor.
      @ In, which, str, o or i, whether outer or inner working dir is needed
      @ Out, working_dir, str, relevant working dir
    """
    if which == 'outer':
      io = 'o'
    elif which == 'inner':
      io = 'i'
    else:
      raise NotImplementedError(f'Unrecognized working dir request: "{which}"')
    return f'{self.name}_{io}'

  def load_econ(self, components):
    """
      Loads active component cashflows
      @ In, components, list, list of HERON components
      @ Out, None
    """
    if 'active' not in self._global_econ:
      # NOTE self._metric can only be NPV right now!
      ## so no need for the "target" in the indicator
      indic = {'name': [self._metric]}
      indic['active'] = []
      for comp in components:
        comp_name = comp.name
        for cf in comp.get_cashflows():
          cf_name = cf.name
          indic['active'].append(f'{comp_name}|{cf_name}')
      self._global_econ['Indicator'] = indic

  def get_econ(self, components):
    """
      Accessor for economic settings for this case
      @ In, components, list, list of HERON components
      @ Out, get_econ, dict, dictionary of global economic settings
    """
    self.load_econ(components)
    return self._global_econ

  def get_labels(self):
    """
      Accessor
      @ In, None
      @ Out, _labels, dict, labels for this case
    """
    return self._labels

  def get_metric(self):
    """
      Accessor
      @ In, None
      @ Out, metric, str, target metric for this case
    """
    return self._metric

  def get_mode(self):
    """
      Accessor
      @ In, None
      @ Out, mode, str, mode of analysis for this case (sweep, opt, etc)
    """
    return self._mode

  def get_verbosity(self):
    """
      Accessor
      @ In, None
      @ Out, verbosity, str, level of vebosity for RAVEN inner/outer runs.
    """
    return self._verbosity

  def get_num_samples(self):
    """
      Accessor
      @ In, None
      @ Out, num_samples, int, number of dispatch realizations to consider
    """
    if self.debug['enabled']:
      return self.debug['inner_samples']
    else:
      return self._num_samples

  def get_num_timesteps(self):
    """
      Accessor
      @ In, None
      @ Out, num_timesteps, int, number of time steps for inner dispatch
    """
    return self._num_hist

  def get_time_name(self):
    """
      Provides the name of the time variable.
      @ In, None
      @ Out, time name, string, name of time variable
    """
    return self._time_varname

  def get_year_name(self):
    """
      Provides the name of the time variable.
      @ In, None
      @ Out, time name, string, name of time variable
    """
    return self._year_varname

  def get_Resample_T(self):
    """
      Accessor
      @ In, None
      @ Out, Resample_T, float, user-requested time deltas
    """
    return self._Resample_T

  def get_hist_interval(self):
    """
      Accessor
      @ In, None
      @ Out, hist_interval, float, user-requested time deltas
    """
    return self._hist_interval

  def get_hist_length(self):
    """
      Accessor
      @ In, None
      @ Out, hist_len, int, length of inner histories
    """
    return self._hist_len

  def get_dispatch_var(self, name):
    """
      Accessor
      @ In, name, str, the name of the dispatch_var
      @ Out, dispatch_var, ValuedParamHandler, a ValuedParam object.
    """
    return self.dispatch_vars[name]

  #### API ####
  def write_workflows(self, components, sources, loc):
    """
      Writes workflows for this case to XMLs on disk.
      @ In, components, HERON components, components for the simulation
      @ In, sources, HERON sources, sources for the simulation
      @ In, loc, str, location in which to write files
      @ Out, None
    """
    # load templates
    template_class = self._load_template()
    inner, outer, cash = template_class.createWorkflow(self, components, sources)

    template_class.writeWorkflow((inner, outer, cash), loc)

  #### UTILITIES ####
  def _load_template(self):
    """
      Loads template files for modification
      @ In, None
      @ Out, template_class, RAVEN Template, instantiated Template class
    """
    src_dir = os.path.dirname(os.path.realpath(__file__))
    heron_dir = os.path.abspath(os.path.join(src_dir, '..'))
    template_dir = os.path.abspath(os.path.join(heron_dir, 'templates'))
    template_name = 'template_driver'
    # import template module
    sys.path.append(heron_dir)
    module = importlib.import_module(f'templates.{template_name}', package="HERON")
    # load template, perform actions
    template_class = module.Template(messageHandler=self.messageHandler)
    template_class.loadTemplate(template_dir)
    return template_class



# Copyright 2022, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
A script to autoload information from HYBRID to HERON.

This script autoloads the needed economic information about a specific component or an element of the grid system from HYBRID text files to a HERON input xml file.
It takes the following command line argument:
1. HERON_input file: the initial HERON xml file (pre_heron_input) before loading any information from HYRBID
To run the script, use the following command:
python hybrid2heron_economic.py path/to/HERON/xml/file
for example, the command looks like this:
python hybrid2heron_economic.py pre_heron_input.xml
The inital HERON xml file (e.g. pre_heron_input.xml) must have:
1- "Components" node : <Components>
2- At least one "Component" subnode under the "Components" node:
  <Component name="component_name">
3- An empty "economics" node under the "component" node with "src" attribute. For example: <economics src="path/to/HYBRID/file"></economics>
"""

#####
# Section 0
# Importing libraries and modules

import os
import sys
import csv
import argparse
from xml.etree import ElementTree as ET

#Note: xm only uses prettify which is in both xmlUtils and convert_utils
try:
  from ravenframework.utils import xmlUtils as xm
  import HERON.src
  HERON_src_path = HERON.src.__path__[0]
except ModuleNotFoundError:
  # # Importing XML utility from RAVEN to make the output XML file looks pretty
  HERON_src_path = os.path.dirname(os.path.abspath(__file__)).split("HERON")[0] + "HERON/src"
  sys.path.append(HERON_src_path)
  from _utils import get_raven_loc
  sys.path.append(get_raven_loc())
  from scripts.conversionScripts import convert_utils as xm

#####
# Section 1: HYBRID Information Extraction Function


def relevant_info_from_hybrid(filepath):
  """
    Extracts the HYBRID keywords, corresponding values and comments.
    @ In, filepath, string, the path of the HYBRID text file that contains the component economic information
    @ Out, all_hybrid_info, dict, the relevant information from the HYBRID text files which includes: variables and their values and comments plus warning messages if necessary
  """
  hybrid_vars, hybrid_values, hybrid_comments = [], [], []
  with open(filepath, encoding="utf8") as f:
    for line in f:
      # if it is not a comment line
      if not line.lstrip().startswith("#"):
        if "=" in line:  # extracting relevant lines from HYBRID files
          relevant_lines = ("".join(line.split("#")[0])).split("=")
          if len(line.split("#")) == 2:  # if comments exist
            the_comment = ("".join(line.split("#")[1])).rstrip()
          else:  # if no comments exist
            the_comment = ""
          if len(line.split("=")) == 2:
            if "#" not in line.split("=")[0]:
              variable_name = relevant_lines[0].strip()
              value = relevant_lines[1].strip()
          if variable_name in Hybrid_keywords:
            hybrid_vars.append(variable_name)
            hybrid_values.append(value)
            hybrid_comments.append(the_comment)
          else:
            print(
              "\n",
              f" The variable {variable_name} in the file '{filepath}' is irrelevant",
              )

  # The Hybrid keywords that, if present, require adding more variables to the HERON input XML file to make it look complete. Since these variables, are not provided by HYBRID, they are assigned temporary default values

  demanding_hybrid_keywords = list(set(hybrid_vars) & set(Required_if))
  extra_variables, extra_values, extra_comments = [], [], []
  for i in range(len(Hybrid_keywords)):
    # for a hybrid keyword that is not already provided by HYBRID but need to be inlucded to make the HERON input file look complete
    if Hybrid_keywords[i] not in hybrid_vars:
      if Required_if[i] in demanding_hybrid_keywords:
        extra_variables.append(Hybrid_keywords[i])
        extra_variable_value = default[i]
        extra_values.append(extra_variable_value)
        extra_comments.append(
          f"Warning: The value of '{Hybrid_keywords[i]}' is not provided by HYBRID. A default value is assigned instead. It needs to be reviewed or replaced"
          )
  all_hybrid_info = dict()
  total_hybrid_variables = hybrid_vars + extra_variables
  total_hybrid_values = hybrid_values + extra_values
  total_hybrid_comments = hybrid_comments + extra_comments
  all_hybrid_info["variables"] = total_hybrid_variables
  all_hybrid_info["values"] = total_hybrid_values
  all_hybrid_info["comments"] = total_hybrid_comments
  if extra_variables:
    all_hybrid_info["warning"] = (
      "\nWarning: The folllowing list of variables were not provided by the HYBRID file:"
      + f"'{filepath}' \nand are assigned default values. The user needs to review them"
      + f" \n{extra_variables}"
    )

  else:
    all_hybrid_info["warning"] = (
      "\n",
      f" All variables were provided by the HYBRID file '{filepath}' and no default values are assigned",
    )

  return all_hybrid_info

#####
# Section 2: A function to covert HYBRID keywords to HERON nodes
# Loading the information from the HYRBID files to the corresopnding nodes, subnodes, parameters at the HERON XML file


def heron_node_from_hybrid(economics_node, hybrid_var_list, value_list, comments_list):
  """
    Creates HERON nodes from HYBRID keywords.
    @ In, economics_node, xml.etree.ElementTree.Element, the "economics" node in the HERON input xml file
    @ In, HYBRID_var_list, list, a list of the relevant HYBRID variables in the text file
    @ In, value_list, list, a list of the relevant HYBRID variables' values in the text file
    @ In, comments_list, list, a list of the relevant HYBRID variables' comments in the text file
    @ Out, economics_node, xml.etree.ElementTree.Element, the "economics" node in the HERON input xml file after being filled with nodes that are constructed based on the HYBRID text files information
  """

  # Initializing lists for HYRBID, HERON keywords, parameters,.etc to load information from the HYBRID text files
  (
    node_or_param_file,
    node_file,
    subnode_file,
    sub_subnode_file,
    belong_to_same_node_file,
    node_index,
  ) = ([], [], [], [], [], [])
  for var in hybrid_var_list:
    (
      node_or_parameter,
      nodes,
      subnodes,
      sub_subnodes,
      belonging_to_same_node,
    ) = hybrid_heron_dict[var][0:5]
    node_or_param_file.append(node_or_parameter)
    node_file.append(nodes)
    subnode_file.append(subnodes)
    sub_subnode_file.append(sub_subnodes)
    belong_to_same_node_file.append(belonging_to_same_node)
    node_index.append(belonging_to_same_node.split(","))

  # Computing the number of HERON XML file nodes under the  "economics"
  maximum_node_index = max([int(item) for sublist in node_index for item in sublist])
  try:
    for node_number in range(maximum_node_index + 1):  # Iterate over the nodes
      main_nodes, parameters = [], []
      nodes_of_same_param, nodes_of_same_subnodes, nodes_of_same_sub_subnodes = [], [], []
      subnodes_of_same_node, sub_subnodes_of_same_node, subnode_of_same_sub_subnodes = [], [], []
      nodes_values, param_values, subnodes_values, subsubnodes_values = [], [], [], []
      nodes_comments, param_comments, subnodes_comments, subsubnodes_comments = [], [], [], []

      for i in range(len(hybrid_var_list)):  # Iterate over HYBRID variables
        # Iterate over correspodning HERON keywords of the same node
        if str(node_number) in belong_to_same_node_file[i].split(","):

          if node_file[i] and not subnode_file[i]:  # nodes with no subnodes
            if node_or_param_file[i] == "N":
              main_nodes.append(node_file[i])
              nodes_values.append(value_list[i])
              nodes_comments.append(comments_list[i])

          if node_file[i] and subnode_file[i]:  # nodes with subnodes
            if sub_subnode_file [i]:  # subsubnode under a subnode under a node
              if node_or_param_file[i] == "N":
                nodes_of_same_sub_subnodes.append(node_file[i])
                subnode_of_same_sub_subnodes.append(subnode_file[i])
                sub_subnodes_of_same_node.append(sub_subnode_file[i])
                subsubnodes_values.append(value_list[i])
                subsubnodes_comments.append(comments_list[i])

            if not sub_subnode_file[i]:
              if node_or_param_file[i] == "P":  # parameters
                nodes_of_same_param.append(node_file[i])
                parameters.append(subnode_file[i])
                param_values.append(value_list[i])
                param_comments.append(comments_list[i])

              if node_or_param_file[i] == "N":  # nodes not parameters
                nodes_of_same_subnodes.append(node_file[i])
                subnodes_of_same_node.append(subnode_file[i])
                subnodes_values.append(value_list[i])
                subnodes_comments.append(comments_list[i])

      if main_nodes:
        subnode0 = ET.SubElement(economics_node, main_nodes[0])
        subnode0.text = nodes_values[0]
        if nodes_comments[0] != "":
          subnode0.append(ET.Comment(nodes_comments[0]))

      param_dict = dict(zip(parameters, param_values))
      if nodes_of_same_param:
        subnode1 = ET.SubElement(
          economics_node, nodes_of_same_param[0], param_dict
          )
        for comment in param_comments:
          parameter_comment = comment
          subnode1.append(ET.Comment(parameter_comment))
      else:
        if nodes_of_same_subnodes:
          subnode1 = ET.SubElement(economics_node, nodes_of_same_subnodes[0])
        else:
          if nodes_of_same_sub_subnodes:
            subnode1 = ET.SubElement(
              economics_node, nodes_of_same_sub_subnodes[0]
              )

      for n in range(len(subnodes_of_same_node)):
        subnode2 = ET.SubElement(subnode1, subnodes_of_same_node[n])
        subnode2.text = str(subnodes_values[n])
        if subnodes_comments[n] != "":
          subnode2.append(ET.Comment(subnodes_comments[n]))
      for m in range(len(subnode_of_same_sub_subnodes)):
        subnode2 = ET.SubElement(subnode1, subnode_of_same_sub_subnodes[m])
        if subsubnodes_comments[m] != "":
          subnode2.append(ET.Comment(subsubnodes_comments[m]))
        subnode3 = ET.SubElement(subnode2, sub_subnodes_of_same_node[m])
        subnode3.text = str(subsubnodes_values[m])


  except IndexError:
    pass

  return economics_node

#####
# Section 3: Connecting HYBRID and HERON keywords
# Load the HYBRID keywords from the csv file
# If needed, these HYBRID keywords can be modified in the csv file
keywords_path = HERON_src_path + "/Hybrid2Heron/HYBRID_HERON_keywords.csv"
with open(keywords_path, "r", encoding="utf8") as csv_file:
  csv_reader = csv.reader(csv_file, delimiter=",")
  next(csv_reader)
  Hybrid_keywords = []  # Expected keywords of HYRBID

  # Is this HERON keywors a node or a parementer in the HERON XML file?
  # N: node, P, parameter
  HERON_Node_or_Parameter = []

  # The location of the HERON node corresponding to the HYBRID keyword is identified by HERON_Node/HERON_subnode/HERON_subsubnode as follows:
  # The main nodes under the "economics" node such as the lifetime and the cashflows
  HERON_Node = []
  HERON_Subnode, HERON_subsubnode = [], []

  # list of subnodes or pararmeters that are attached to the same node
  belong_to_same_node = []

  # Variables that are required to be included, even if not provided by HYBRID, because of the presence of other variables (To make the HERON xml input file complete)
  Required_if = []

  # default values for variables that are not provided but must be included
  default = []

  for lines in csv_reader:
    Hybrid_keywords.append(lines[0])
    HERON_Node_or_Parameter.append(lines[1])
    HERON_Node.append(lines[2])
    HERON_Subnode.append(lines[3])
    HERON_subsubnode.append(lines[4])
    belong_to_same_node.append(lines[5])
    Required_if.append(lines[6])
    default.append(lines[7])

hybrid_heron_dict = {
  i: [j, k, l, m, n, o, p]
  for i, j, k, l, m, n, o, p in zip(
  Hybrid_keywords,
  HERON_Node_or_Parameter,
  HERON_Node,
  HERON_Subnode,
  HERON_subsubnode,
  belong_to_same_node,
  Required_if,
  default,
  )
}

#####
# Section 4
# Specifying user inputs and output file
if __name__ == "__main__":
  parser = argparse.ArgumentParser(
    description="Autoload the component economic information from HYBRID folder to HERON input XML file"
    )
    # Inputs from the user
  parser.add_argument("HERON_input_file", help="HERON XML input file path")
  args = parser.parse_args()
  # Output file where you get the new HYBRID xml file with the new information loaded from HYBRID
  output_file = "heron_input.xml"

####
# Section 5: Autoload the HERON XML input file uing the HYBRID information
# Reading the HERON xml file
# Finding the "Compnents" node and the "economics" node under each component to fill it from HYBRID text files

# The HERON XML file tree
HERON_inp_tree = ET.parse(args.HERON_input_file)
components_list = HERON_inp_tree.findall("Components")  # The "components" node
if not components_list:
  print("\n", "The 'Components' node is not found in the HERON input xml file")

# Searching for the node "Component"
for components in components_list:
  component = components.findall("Component")
  comp_list = []  # The list of compoenents
  for comp in component:
    comp_list.append(comp.attrib["name"])
    print(f"\nThe component '{comp.attrib['name']}' is found")
    for node in comp:
      if node.tag == "economics":
        print(
          f"The 'economics' node is found at the component '{comp.attrib['name']}'"
          )
        ECO_NODE_FOUND = "True"
        if node.attrib["src"]:
          print("The attribute 'src' is found at the 'economics' node")
          children_nodes = []
          for child in node:
            children_nodes.append(child)
          if children_nodes:
              print(
                "\n",
                f"Warning: The 'economics' node at the component '{comp.attrib['name']}' is not empty so it will be left unchanged",
                )
          else:
            print(
              f"The economic information for the component '{comp.attrib['name']}' is loaded from the file: '{node.attrib['src']}'"
              )
            # check whether the file path exists
            if os.path.exists(node.attrib["src"]):
              print(f"The file path '{node.attrib['src']}' exists")
              comp.remove(node)
              print(
              "The 'economics' node in the"
              f" component '{comp.attrib['name']}' is being modified",
              )

              # If HYBRID variables list is not empty, start filling the "economics" node
              if relevant_info_from_hybrid(node.attrib["src"])["variables"]:
                node_eco = ET.SubElement(comp, "economics")
                node_eco.append(
                  ET.Comment(
                    "This 'economics' subnode is created using information from the HYBRID simulations"
                    )
                  )
                inputs = relevant_info_from_hybrid(node.attrib["src"])
                input_vars, input_vals, input_comments = (
                  inputs["variables"],
                  inputs["values"],
                  inputs["comments"],
                )
                print(inputs["warning"])

                # Applying the Hybrid2Heron function
                heron_node_from_hybrid(
                  node_eco, input_vars, input_vals, input_comments
                  )

              else:
                print(
                  f"no relevant information were found at the file {node.attrib['src']}'"
                  )

            else:
              print(
                f"Warning: The file path '{node.attrib['src']}' does not exist"
                )

        else:
          print(
            "The attribute 'src' is not found at the 'economics' node so it will be left unchanged"
            )

    if not ECO_NODE_FOUND:
      print(f"The economics node is not found at the component{comp}")

  if not comp_list:
    print("No componenet names were found")

#####
# Section 6: Print the modified HERON input new file in a pretty RAVEN-preferred format
with open(output_file, "w", encoding="utf8") as out:
  print((xm.prettify(HERON_inp_tree)), file=out)



#!/usr/bin/env python
# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Runs HERON.
"""
import os
from queue import Empty
import sys
import argparse

sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..'))
import HERON.src._utils as hutils
try:
  import ravenframework
except ModuleNotFoundError:
  sys.path.append(hutils.get_raven_loc())

from HERON.src import input_loader
from HERON.src.base import Base
from HERON.src.Moped import MOPED
from HERON.src import Herd

from ravenframework.MessageHandler import MessageHandler


class HERON(Base):
  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    Base.__init__(self)
    self._components = []   # units involved in this study
    self._sources = []      # sources involved in this study (arma, functions, static histories, etc)
    self._case = None       # information about the case we're running
    self._input_dir = None  # location of the input XML file
    self._input_name = None # name of the input XML file

    messageHandler = MessageHandler()
    messageHandler.initialize({'verbosity': 'all',
                               'callerLength': 18,
                               'tagLength': 7,
                               'suppressErrs': False,})
    self.messageHandler = messageHandler

  def read_input(self, name):
    """
      Loads data from input
      @ In, name, str, name of file to read from
      @ Out, None
    """
    location, fname = os.path.split(name)
    self._input_dir = location
    self._input_name = fname
    inp = input_loader.load(name)
    objects = input_loader.parse(inp, location, self.messageHandler)
    self._components = objects['components']
    self._sources = objects['sources']
    self._case = objects['case']

  def __repr__(self):
    """
      String representation of object.
      @ In, None
      @ Out, repr, str, string rep
    """
    return '<HERON Simulation>'

  def print_me(self, tabs=0, tab='  '):
    """
      Prints info about self.
      @ In, tabs, int, number of tabs to insert
      @ In, tab, str, how tab should be written
      @ Out, None
    """
    pre = tab*tabs
    self.raiseADebug("==========================")
    self.raiseADebug("Printing simulation state:")
    self.raiseADebug("==========================")
    assert self._case is not None
    self._case.print_me(tabs=tabs+1, tab=tab)
    for comp in self._components:
      comp.print_me(tabs=tabs+1, tab=tab)
    for source in self._sources:
      source.print_me(tabs=tabs+1, tab=tab)

  def create_raven_workflow(self, case=None):
    """
      Loads, modifies, and writes a RAVEN template workflow based on the Case.
      @ In, case, Cases.Case, optional, case to run (defaults to self._case)
      @ Out, None
    """
    if case is None:
      case = self._case
    # let the case do the work
    assert case is not None
    case.write_workflows(self._components, self._sources, self._input_dir)

  def run_moped_workflow(self, case=None, components=None, sources=None):
    """
      Runs MOPED workflow for generating pyomo problem and solves it
      @ In, case, HERON case object with necessary run settings
      @ Out, None
    """
    if case is None:
      case = self._case
    if components is None:
      components = self._components
    if sources is None:
      sources = self._sources
    assert case is not None and components is not None and sources is not None
    moped = MOPED()
    self.raiseAMessage("***** You are running Monolithic Optimizer for Probabilistic Economic Dispatch (MOPED) *****")
    moped.setInitialParams(case, components, sources)
    moped.run()

  def run_dispatches_workflow(self):
    """
      Runs DISPATCHES workflow for creating framework and running with IDAES
      @ In, None
      @ Out, None
    """
    case = self._case
    components = self._components
    sources = self._sources
    assert case is not None and components is not None and sources is not None
    dispatches = Herd.HERD()
    print("*******************************************************************************")
    print("HERON is Running DISPATCHES")
    print("*******************************************************************************")
    dispatches.setInitialParams(case, components, sources)
    dispatches.run()

def main():
  """
    Runs HERON input from command line arguments
    @ In, None
    @ Out, None
  """
  parser = argparse.ArgumentParser(description='Holistic Energy Resource Optimization Network (HERON)')
  parser.add_argument('xml_input_file', help='HERON XML input file')
  args = parser.parse_args()
  sim = HERON()
  sim.read_input(args.xml_input_file) # TODO expand to use arguments?
  # print details
  sim.print_me()
  if sim._case._workflow == 'standard':
    sim.create_raven_workflow()
  elif sim._case._workflow == 'MOPED':
    sim.run_moped_workflow()
  elif sim._case._workflow == 'DISPATCHES':
    sim.run_dispatches_workflow()
  # TODO someday? sim.run()


if __name__ == '__main__':
  main()

